from __future__ import division, print_function

from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy_utils import create_database, drop_database

from beam_nuggets.io.relational_db_api import (
    create_table,
    load_table,
    get_column_names_from_table,
    TableConfiguration,
)


class TestDatabase(object):
    def __init__(self, source_config):
        self._url = source_config.url
        self._SessionClass = sessionmaker(bind=create_engine(self._url))

    def init_db(self):
        create_database(self._url)

    def destroy_db(self):
        drop_database(self._url)  # will also remove the sqlite file

    def create_table(self, name, create_table_if_missing, define_table_f):
        # create the table
        with self.session_scope() as session:
            create_table(
                session=session,
                name=name,
                table_config=TableConfiguration(
                    name=name,
                    define_table_f=define_table_f,
                    create_if_missing=create_table_if_missing
                ),
                record=None
            )

    def load_table_class(self, table_name):
        with self.session_scope() as session:
            return load_table(session, table_name)

    def write_rows(self, table_name, rows):
        with self.session_scope() as session:
            TableCls = load_table(session, table_name)
            session.add_all([TableCls(**row_dict) for row_dict in rows])

    def read_rows(self, table_name):
        with self.session_scope() as session:
            TableCls = load_table(session, table_name)
            if TableCls:
                column_names = get_column_names_from_table(TableCls)

                def to_dict(db_row):
                    return {col: getattr(db_row, col) for col in column_names}

                rows = [to_dict(db_row) for db_row in session.query(TableCls)]
            else:
                rows = []
        return rows

    @contextmanager
    def session_scope(self):
        session = self._SessionClass()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close_all()
            session.bind.dispose()



from __future__ import division, print_function

import datetime
import unittest

import apache_beam as beam
from apache_beam.testing.test_pipeline import TestPipeline
from nose.tools import assert_equal, assert_not_equal
from sqlalchemy import DateTime

from beam_nuggets.io import relational_db
from .test_base import TransformBaseTest


def sort(ls, key):
    return sorted(ls, key=lambda item: item[key])


class TestWriteTransform(TransformBaseTest):

    def setUp(self):
        super(TestWriteTransform, self).setUp()
        self.records, self.table_name = self.get_test_records_and_table_name()

    def execute_pipeline(self, source_config, table_config, records):
        with TestPipeline() as p:
            months = p | "Reading records" >> beam.Create(records)
            months | 'Writing to table' >> relational_db.Write(
                source_config=source_config,
                table_config=table_config
            )

        # retrieve the table rows
        return self.db.read_rows(table_config.name)

    def test_write(self):
        table_config = relational_db.TableConfiguration(
            name=self.table_name,
            create_if_missing=True,
            primary_key_columns=['num']
        )

        # execute the write pipeline and retrieve the table rows
        table_rows = self.execute_pipeline(
            source_config=self.source_config,
            table_config=table_config,
            records=self.records
        )

        # compare
        assert_equal(sort(table_rows, 'num'), sort(self.records, 'num'))

    def test_write_no_primary_key(self):
        table_config = relational_db.TableConfiguration(
            name=self.table_name,
            create_if_missing=True
        )

        # execute the write pipeline and retrieve the table rows
        table_rows = self.execute_pipeline(
            source_config=self.source_config,
            table_config=table_config,
            records=self.records
        )

        # An auto-increment table called id should have been added
        expected_table_rows = [dict(record) for record in self.records]
        _ = [rec.update(id=i + 1) for i, rec in enumerate(expected_table_rows)]

        assert_equal(sort(table_rows, 'num'), sort(expected_table_rows, 'num'))

    def test_write_to_user_defined_table(self):
        user_defined_table = 'my_table'
        assert_not_equal(user_defined_table, self.table_name)

        def define_table(metadata):
            from sqlalchemy import Table, Column, Integer, String
            return Table(
                user_defined_table, metadata,
                Column('name', String(50), primary_key=True),
                Column('num', Integer)
            )

        table_config = relational_db.TableConfiguration(
            name=user_defined_table,
            define_table_f=define_table,
            create_if_missing=True,
        )

        # execute the write pipeline and retrieve the table rows
        table_rows = self.execute_pipeline(
            source_config=self.source_config,
            table_config=table_config,
            records=self.records
        )

        # assert nothing written to the test default table
        assert_equal(len(self.db.read_rows(self.table_name)), 0)

        # and all written to the user defined table
        assert_equal(sort(table_rows, 'name'), sort(self.records, 'name'))

    def test_write_to_existing_table(self):
        table_name = self.create_table()

        table_config = relational_db.TableConfiguration(
            name=table_name,
            create_if_missing=False
        )

        part1_size = 2
        part1_records = self.records[:part1_size]
        part2_records = self.records[part1_size:]

        # write part1 to the DB
        part1_table_rows = self.execute_pipeline(
            source_config=self.source_config,
            table_config=table_config,
            records=part1_records
        )
        assert_equal(sort(part1_table_rows, 'num'), sort(part1_records, 'num'))

        # write part2 to the DB
        part2_table_rows = self.execute_pipeline(
            source_config=self.source_config,
            table_config=table_config,
            records=part2_records
        )
        assert_equal(
            sort(part2_table_rows, 'num'),
            sort(part1_records + part2_records, 'num')
        )
        # Note, above assumes row are returned in the same order as they were
        # written (i.e. first written first returned)

    def test_auto_column_type_inference(self):
        import pandas as pd
        from sqlalchemy import Integer, Float, String
        NAME = 'name'
        NUM = 'num'
        TIME_STAMP = 'time_stamp'
        records = [
            {NAME: name, NUM: num, TIME_STAMP: ts}
            for name, num, ts in [
                ['Jan', 1, datetime.datetime.now()],
                ['Feb', 2, datetime.datetime.now()]
            ]
        ]
        expected_column_types = [
            (NAME, String),
            (NUM, Float),
            (TIME_STAMP, DateTime),
            ('id', Integer)  # Auto created  when no primary key is specified
        ]

        table_config = relational_db.TableConfiguration(
            name=self.table_name,
            create_if_missing=True
        )

        # write the records using the pipeline
        _ = self.execute_pipeline(
            source_config=self.source_config,
            table_config=table_config,
            records=records
        )

        # load created table metadata
        table = self.db.load_table_class(self.table_name).__table__
        columns = [col for col in table.columns]
        get_column = lambda name: filter(lambda c: c.name == name, columns)[0]

        # verify inferred column types is as expected
        for col_name, expec_col_type in expected_column_types:
            inferred_col_type = get_column(col_name).type
            assert_equal(isinstance(inferred_col_type, expec_col_type), True)

    def test_write_to_postgres_with_upsert(self):
        if not self.postgres_source_config:
            raise unittest.SkipTest(
                '"{}" should run against postgres instance'.format(self.id())
            )
        self.evaluate_upsert(self.postgres_source_config)

    def test_write_to_mysql_with_upsert(self):
        if not self.mysql_source_config:
            raise unittest.SkipTest(
                '"{}" should run against mysql instance'.format(self.id())
            )
        self.evaluate_upsert(self.mysql_source_config)

    def evaluate_upsert(self, source_config):
        # reconfigure the db for this TC (as the db configured in setup
        # used the default source config, which might be different from the
        # one used in this TC).
        self.destroy_db(self.db)
        self.db = self.configure_db(source_config)


        table_config = relational_db.TableConfiguration(
            name=self.table_name,
            create_if_missing=True,
            primary_key_columns=['num']
        )

        # execute the write pipeline and retrieve the table rows
        table_rows = self.execute_pipeline(
            source_config=source_config,
            table_config=table_config,
            records=self.records
        )

        # compare
        assert_equal(table_rows, self.records)

        # re-write the same records
        table_rows_after_rewrite = self.execute_pipeline(
            source_config=source_config,
            table_config=table_config,
            records=self.records
        )

        # verify the table records are the same
        assert_equal(table_rows_after_rewrite, table_rows)

    def create_table(self):
        table_name = 'months_table'

        def define_table(metadata):
            from sqlalchemy import Table, Column, Integer, String
            return Table(
                table_name, metadata,
                Column('name', String(50), primary_key=True),
                Column('num', Integer)
            )

        self.db.create_table(
            name=table_name,
            define_table_f=define_table,
            create_table_if_missing=True
        )

        return table_name

    @staticmethod
    def get_test_records_and_table_name():
        table_name = 'months'
        records = [
            {'name': 'Jan', 'num': 1},
            {'name': 'Feb', 'num': 2},
            {'name': 'Mar', 'num': 3},
            {'name': 'Apr', 'num': 4},
            {'name': 'May', 'num': 5},
        ]
        return records, table_name


if __name__ == '__main__':
    unittest.main()



import unittest

from apache_beam.testing.test_pipeline import TestPipeline
import apache_beam as beam

from beam_nuggets.io import kafkaio


@unittest.skip("tests aren't functional yet")
class TestKafkaProduceTransform(unittest.TestCase):

    def setUp(self):
        super(TestKafkaProduceTransform, self).setUp()

    def test_ProduceKafkaMessage(self):
        #create messages and push messages into Apache Kafka
        with TestPipeline() as p:
            (p | "Creating records" >> beam.Create([('dev_1', '{"device": "0001", status": "healthy"}')])
               | "Produce kafka message" >> kafkaio.KafkaProduce(topic="test_stream", servers="localhost:9092")
                )


@unittest.skip("tests aren't functional yet")
class TestKafkaConsumeTransform(unittest.TestCase):

    def setUp(self):
        super(TestKafkaConsumeTransform, self).setUp()

    def test_ConsumeFromKafka(self):
        kafka_config = {"topic": "test_stream",
                        "bootstrap_servers": "localhost:9092",
                        "group_id": "test_group"}
        #create a streaming Kafka consumer
        with TestPipeline() as p:
            p | "Consume kafka messages" >> kafkaio.KafkaConsume(kafka_config)


if __name__ == '__main__':
    unittest.main()



from __future__ import division, print_function

import datetime
import unittest

import numpy as np
from nose.tools import assert_equal
from sqlalchemy import (
    String,
    Float,
    Boolean,
    DateTime,
    Date
)

from beam_nuggets.io.relational_db_api import (
    infer_db_type,
    VARCHAR_DEFAULT_COL_SIZE
)


class RelationalDBAPITest(unittest.TestCase):
    # PostgreSQL and SQLite support String columns with no max size
    postgres_sqlite_str_value_to_expected_db_type = [
        ('sss', String),
        (None, String),
    ]

    str_value_to_expected_db_type = [
        ('sss', String(VARCHAR_DEFAULT_COL_SIZE)),
        (None, String(VARCHAR_DEFAULT_COL_SIZE)),
    ]

    other_types_values_to_db_type = [
        (10, Float),
        (10.5, Float),
        (np.nan, Float),
        (True, Boolean),
        (datetime.date.today(), Date),
        (datetime.datetime.today(), DateTime),
    ]

    def validate_infer_db_type(self, value_to_expected_db_type, drivername):
        for value, expected_db_type in value_to_expected_db_type:
            inferred_type = infer_db_type(value, drivername)
            assert_equal(
                expected_db_type,
                inferred_type,
                'expected: {}, got: {}'.format(
                    type(expected_db_type),
                    type(inferred_type)
                )
            )

    def test_infer_db_type_postgres(self):
        self.validate_infer_db_type(
            self.postgres_sqlite_str_value_to_expected_db_type +
            self.other_types_values_to_db_type,
            'postgresql'
        )

    def test_infer_db_type_sqlite(self):
        self.validate_infer_db_type(
            self.postgres_sqlite_str_value_to_expected_db_type +
            self.other_types_values_to_db_type,
            'sqlite'
        )

    def test_infer_db_type_mysql(self):
        drivername = 'mysql'

        self.validate_infer_db_type(
            self.other_types_values_to_db_type,
            drivername
        )

        # For some reason, I didn't chase, for fixed String size columns
        # inferred_type == expected_db_type is False even if the two seemed
        # equal. Below works enough.
        for value, expected_db_type in \
            self.str_value_to_expected_db_type:
            inferred_type = infer_db_type(value, drivername)
            assert_equal(
                type(expected_db_type),
                type(inferred_type),
                'expected: {}, got: {}'.format(
                    type(expected_db_type),
                    type(inferred_type)
                )
            )


if __name__ == '__main__':
    unittest.main()
